# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow/core/protobuf/tpu/optimization_parameters.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from tensorflow.compiler.xla.service import hlo_pb2 as tensorflow_dot_compiler_dot_xla_dot_service_dot_hlo__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorflow/core/protobuf/tpu/optimization_parameters.proto',
  package='tensorflow.tpu',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n:tensorflow/core/protobuf/tpu/optimization_parameters.proto\x12\x0etensorflow.tpu\x1a\x1egoogle/protobuf/wrappers.proto\x1a)tensorflow/compiler/xla/service/hlo.proto\"h\n\x0e\x43lippingLimits\x12*\n\x05lower\x18\x01 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12*\n\x05upper\x18\x02 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\"\"\n\x13\x44ynamicLearningRate\x12\x0b\n\x03tag\x18\x01 \x01(\x05\"k\n\x0cLearningRate\x12\x12\n\x08\x63onstant\x18\x01 \x01(\x02H\x00\x12\x36\n\x07\x64ynamic\x18\x02 \x01(\x0b\x32#.tensorflow.tpu.DynamicLearningRateH\x00\x42\x0f\n\rlearning_rate\".\n\x11\x41\x64\x61gradParametersJ\x04\x08\x01\x10\x02R\x13initial_accumulator\"m\n\x18\x42oundedAdagradParameters\x12 \n\x18update_accumulator_first\x18\x01 \x01(\x08\x12\x16\n\x0emax_var_update\x18\x02 \x01(\x02\x12\x17\n\x0fmax_accumulator\x18\x03 \x01(\x02\"%\n#StochasticGradientDescentParameters\"\xb2\x01\n\x0e\x46trlParameters\x12\n\n\x02l1\x18\x01 \x01(\x02\x12\n\n\x02l2\x18\x02 \x01(\x02\x12\x10\n\x08lr_power\x18\x03 \x01(\x02\x12\x0c\n\x04\x62\x65ta\x18\x07 \x01(\x02\x12\x1d\n\x15multiply_linear_by_lr\x18\x06 \x01(\x08\x12\x1e\n\x16\x61llow_zero_accumulator\x18\x08 \x01(\x08J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06R\rinitial_accumR\x0einitial_linear\"\x99\x01\n\x0e\x41\x64\x61mParameters\x12\r\n\x05\x62\x65ta1\x18\x03 \x01(\x02\x12\r\n\x05\x62\x65ta2\x18\x04 \x01(\x02\x12\x0f\n\x07\x65psilon\x18\x05 \x01(\x02\x12\x19\n\x11use_non_lazy_adam\x18\x08 \x01(\x08\x12\x1b\n\x13use_sum_inside_sqrt\x18\n \x01(\x08J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08R\tinitial_mR\tinitial_v\"Q\n\x12MomentumParameters\x12\x10\n\x08momentum\x18\x01 \x01(\x02\x12\x14\n\x0cuse_nesterov\x18\x02 \x01(\x08J\x04\x08\x03\x10\x04R\rinitial_accum\"h\n\x11RmsPropParameters\x12\x0b\n\x03rho\x18\x01 \x01(\x02\x12\x10\n\x08momentum\x18\x02 \x01(\x02\x12\x0f\n\x07\x65psilon\x18\x03 \x01(\x02J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06R\ninitial_msR\x0binitial_mom\"\x82\x01\n\x19\x43\x65nteredRmsPropParameters\x12\x0b\n\x03rho\x18\x01 \x01(\x02\x12\x10\n\x08momentum\x18\x02 \x01(\x02\x12\x0f\n\x07\x65psilon\x18\x03 \x01(\x02J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06J\x04\x08\x06\x10\x07R\ninitial_msR\x0binitial_momR\ninitial_mg\"\x9d\x03\n\x19MdlAdagradLightParameters\x12\n\n\x02l2\x18\x01 \x01(\x02\x12\x10\n\x08lr_power\x18\x02 \x01(\x02\x12 \n\x18min_servable_mdl_benefit\x18\x03 \x01(\x02\x12\x19\n\x11mdl_mix_in_margin\x18\x04 \x01(\x02\x12 \n\x18mdl_benefit_rampup_coeff\x18\x05 \x01(\x02\x12\x16\n\x0emdl_min_weight\x18\x06 \x01(\x02\x12\x1d\n\x15\x62\x65nefit_revisit_scale\x18\x07 \x01(\x02\x12\x19\n\x11max_event_benefit\x18\x08 \x01(\x02\x12\x19\n\x11max_total_benefit\x18\t \x01(\x02\x12\x16\n\x0emdl_hard_limit\x18\n \x01(\x02\x12\x1e\n\x16hard_limit_min_benefit\x18\x0b \x01(\x08\x12\x16\n\x0emdl_regularize\x18\x0c \x01(\x08J\x04\x08\r\x10\x0eJ\x04\x08\x0e\x10\x0fJ\x04\x08\x0f\x10\x10R\x13initial_accumulatorR\x0einitial_weightR\x0finitial_benefit\"c\n\x12\x41\x64\x61\x64\x65ltaParameters\x12\x0b\n\x03rho\x18\x01 \x01(\x02\x12\x0f\n\x07\x65psilon\x18\x02 \x01(\x02J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05R\x13initial_accumulatorR\x0einitial_update\"N\n\x19ProximalAdagradParameters\x12\n\n\x02l1\x18\x01 \x01(\x02\x12\n\n\x02l2\x18\x02 \x01(\x02J\x04\x08\x03\x10\x04R\x13initial_accumulator\"I\n\x14OnlineYogiParameters\x12\n\n\x02l1\x18\x01 \x01(\x02\x12\n\n\x02l2\x18\x02 \x01(\x02\x12\r\n\x05\x62\x65ta2\x18\x03 \x01(\x02J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08\"k\n\x16ProximalYogiParameters\x12\n\n\x02l1\x18\x01 \x01(\x02\x12\n\n\x02l2\x18\x02 \x01(\x02\x12\r\n\x05\x62\x65ta1\x18\x03 \x01(\x02\x12\r\n\x05\x62\x65ta2\x18\x04 \x01(\x02\x12\x0f\n\x07\x65psilon\x18\x05 \x01(\x02J\x04\x08\x08\x10\tJ\x04\x08\t\x10\n\"r\n\x1c\x46requencyEstimatorParameters\x12\x0b\n\x03tau\x18\x01 \x01(\x02\x12\x11\n\tmax_delta\x18\x02 \x01(\x02\x12\x19\n\x11outlier_threshold\x18\x03 \x01(\x02\x12\x17\n\x0fweight_exponent\x18\x04 \x01(\x02\"\\\n\x1cUserDefinedProgramParameters\x12$\n\x07program\x18\x01 \x01(\x0b\x32\x13.xla.HloModuleProto\x12\x16\n\x0epadding_values\x18\x02 \x03(\x02\"R\n\x1aGradientAccumulationStatus\"4\n\x06Status\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0b\n\x07\x45NABLED\x10\x01\x12\x0c\n\x08\x44ISABLED\x10\x02\"\x9b\x01\n\x1dHotIdReplicationConfiguration\x12\x44\n\x06status\x18\x01 \x01(\x0e\x32\x34.tensorflow.tpu.HotIdReplicationConfiguration.Status\"4\n\x06Status\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0b\n\x07\x45NABLED\x10\x01\x12\x0c\n\x08\x44ISABLED\x10\x02\"\xb6\x0b\n\x16OptimizationParameters\x12\x33\n\rlearning_rate\x18\r \x01(\x0b\x32\x1c.tensorflow.tpu.LearningRate\x12\x37\n\x0f\x63lipping_limits\x18\x02 \x01(\x0b\x32\x1e.tensorflow.tpu.ClippingLimits\x12@\n\x18gradient_clipping_limits\x18\x07 \x01(\x0b\x32\x1e.tensorflow.tpu.ClippingLimits\x12\x1b\n\x13weight_decay_factor\x18\x10 \x01(\x02\x12\x35\n-multiply_weight_decay_factor_by_learning_rate\x18\x16 \x01(\x08\x12W\n\x1cgradient_accumulation_status\x18\x11 \x01(\x0e\x32\x31.tensorflow.tpu.GradientAccumulationStatus.Status\x12W\n hot_id_replication_configuration\x18\x12 \x01(\x0b\x32-.tensorflow.tpu.HotIdReplicationConfiguration\x12\x34\n\x07\x61\x64\x61grad\x18\x03 \x01(\x0b\x32!.tensorflow.tpu.AdagradParametersH\x00\x12\x43\n\x0f\x62ounded_adagrad\x18\x13 \x01(\x0b\x32(.tensorflow.tpu.BoundedAdagradParametersH\x00\x12Z\n\x1bstochastic_gradient_descent\x18\x04 \x01(\x0b\x32\x33.tensorflow.tpu.StochasticGradientDescentParametersH\x00\x12.\n\x04\x66trl\x18\x05 \x01(\x0b\x32\x1e.tensorflow.tpu.FtrlParametersH\x00\x12.\n\x04\x61\x64\x61m\x18\x06 \x01(\x0b\x32\x1e.tensorflow.tpu.AdamParametersH\x00\x12\x36\n\x08momentum\x18\x08 \x01(\x0b\x32\".tensorflow.tpu.MomentumParametersH\x00\x12\x35\n\x08rms_prop\x18\t \x01(\x0b\x32!.tensorflow.tpu.RmsPropParametersH\x00\x12\x46\n\x11\x63\x65ntered_rms_prop\x18\n \x01(\x0b\x32).tensorflow.tpu.CenteredRmsPropParametersH\x00\x12\x46\n\x11mdl_adagrad_light\x18\x0b \x01(\x0b\x32).tensorflow.tpu.MdlAdagradLightParametersH\x00\x12\x36\n\x08\x61\x64\x61\x64\x65lta\x18\x0c \x01(\x0b\x32\".tensorflow.tpu.AdadeltaParametersH\x00\x12\x45\n\x10proximal_adagrad\x18\x0e \x01(\x0b\x32).tensorflow.tpu.ProximalAdagradParametersH\x00\x12;\n\x0bonline_yogi\x18\x14 \x01(\x0b\x32$.tensorflow.tpu.OnlineYogiParametersH\x00\x12?\n\rproximal_yogi\x18\x15 \x01(\x0b\x32&.tensorflow.tpu.ProximalYogiParametersH\x00\x12K\n\x13\x66requency_estimator\x18\x17 \x01(\x0b\x32,.tensorflow.tpu.FrequencyEstimatorParametersH\x00\x12L\n\x14user_defined_program\x18\x18 \x01(\x0b\x32,.tensorflow.tpu.UserDefinedProgramParametersH\x00\x42\x0c\n\nparametersJ\x04\x08\x01\x10\x02J\x04\x08\x0f\x10\x10\"\xb7\x02\n\x1aStateVariableSpecification\x12\x0c\n\x04name\x18\x01 \x01(\t\x12N\n\x0cuser_defined\x18\x02 \x01(\x0b\x32\x36.tensorflow.tpu.StateVariableSpecification.UserDefinedH\x00\x12Y\n\x12\x66ill_with_constant\x18\x03 \x01(\x0b\x32;.tensorflow.tpu.StateVariableSpecification.FillWithConstantH\x00\x1a,\n\x0bUserDefined\x12\x1d\n\x15padding_initial_value\x18\x01 \x01(\x01\x1a)\n\x10\x46illWithConstant\x12\x15\n\rinitial_value\x18\x01 \x01(\x01\x42\x07\n\x05usageb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,tensorflow_dot_compiler_dot_xla_dot_service_dot_hlo__pb2.DESCRIPTOR,])



_GRADIENTACCUMULATIONSTATUS_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='tensorflow.tpu.GradientAccumulationStatus.Status',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENABLED', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DISABLED', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2282,
  serialized_end=2334,
)
_sym_db.RegisterEnumDescriptor(_GRADIENTACCUMULATIONSTATUS_STATUS)

_HOTIDREPLICATIONCONFIGURATION_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='tensorflow.tpu.HotIdReplicationConfiguration.Status',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENABLED', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DISABLED', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2282,
  serialized_end=2334,
)
_sym_db.RegisterEnumDescriptor(_HOTIDREPLICATIONCONFIGURATION_STATUS)


_CLIPPINGLIMITS = _descriptor.Descriptor(
  name='ClippingLimits',
  full_name='tensorflow.tpu.ClippingLimits',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lower', full_name='tensorflow.tpu.ClippingLimits.lower', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='upper', full_name='tensorflow.tpu.ClippingLimits.upper', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=153,
  serialized_end=257,
)


_DYNAMICLEARNINGRATE = _descriptor.Descriptor(
  name='DynamicLearningRate',
  full_name='tensorflow.tpu.DynamicLearningRate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tag', full_name='tensorflow.tpu.DynamicLearningRate.tag', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=259,
  serialized_end=293,
)


_LEARNINGRATE = _descriptor.Descriptor(
  name='LearningRate',
  full_name='tensorflow.tpu.LearningRate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='constant', full_name='tensorflow.tpu.LearningRate.constant', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dynamic', full_name='tensorflow.tpu.LearningRate.dynamic', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='learning_rate', full_name='tensorflow.tpu.LearningRate.learning_rate',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=295,
  serialized_end=402,
)


_ADAGRADPARAMETERS = _descriptor.Descriptor(
  name='AdagradParameters',
  full_name='tensorflow.tpu.AdagradParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=404,
  serialized_end=450,
)


_BOUNDEDADAGRADPARAMETERS = _descriptor.Descriptor(
  name='BoundedAdagradParameters',
  full_name='tensorflow.tpu.BoundedAdagradParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='update_accumulator_first', full_name='tensorflow.tpu.BoundedAdagradParameters.update_accumulator_first', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_var_update', full_name='tensorflow.tpu.BoundedAdagradParameters.max_var_update', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_accumulator', full_name='tensorflow.tpu.BoundedAdagradParameters.max_accumulator', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=452,
  serialized_end=561,
)


_STOCHASTICGRADIENTDESCENTPARAMETERS = _descriptor.Descriptor(
  name='StochasticGradientDescentParameters',
  full_name='tensorflow.tpu.StochasticGradientDescentParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=563,
  serialized_end=600,
)


_FTRLPARAMETERS = _descriptor.Descriptor(
  name='FtrlParameters',
  full_name='tensorflow.tpu.FtrlParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='l1', full_name='tensorflow.tpu.FtrlParameters.l1', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='l2', full_name='tensorflow.tpu.FtrlParameters.l2', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lr_power', full_name='tensorflow.tpu.FtrlParameters.lr_power', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beta', full_name='tensorflow.tpu.FtrlParameters.beta', index=3,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multiply_linear_by_lr', full_name='tensorflow.tpu.FtrlParameters.multiply_linear_by_lr', index=4,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='allow_zero_accumulator', full_name='tensorflow.tpu.FtrlParameters.allow_zero_accumulator', index=5,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=603,
  serialized_end=781,
)


_ADAMPARAMETERS = _descriptor.Descriptor(
  name='AdamParameters',
  full_name='tensorflow.tpu.AdamParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='beta1', full_name='tensorflow.tpu.AdamParameters.beta1', index=0,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beta2', full_name='tensorflow.tpu.AdamParameters.beta2', index=1,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='epsilon', full_name='tensorflow.tpu.AdamParameters.epsilon', index=2,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='use_non_lazy_adam', full_name='tensorflow.tpu.AdamParameters.use_non_lazy_adam', index=3,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='use_sum_inside_sqrt', full_name='tensorflow.tpu.AdamParameters.use_sum_inside_sqrt', index=4,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=784,
  serialized_end=937,
)


_MOMENTUMPARAMETERS = _descriptor.Descriptor(
  name='MomentumParameters',
  full_name='tensorflow.tpu.MomentumParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='momentum', full_name='tensorflow.tpu.MomentumParameters.momentum', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='use_nesterov', full_name='tensorflow.tpu.MomentumParameters.use_nesterov', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=939,
  serialized_end=1020,
)


_RMSPROPPARAMETERS = _descriptor.Descriptor(
  name='RmsPropParameters',
  full_name='tensorflow.tpu.RmsPropParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rho', full_name='tensorflow.tpu.RmsPropParameters.rho', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='momentum', full_name='tensorflow.tpu.RmsPropParameters.momentum', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='epsilon', full_name='tensorflow.tpu.RmsPropParameters.epsilon', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1022,
  serialized_end=1126,
)


_CENTEREDRMSPROPPARAMETERS = _descriptor.Descriptor(
  name='CenteredRmsPropParameters',
  full_name='tensorflow.tpu.CenteredRmsPropParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rho', full_name='tensorflow.tpu.CenteredRmsPropParameters.rho', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='momentum', full_name='tensorflow.tpu.CenteredRmsPropParameters.momentum', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='epsilon', full_name='tensorflow.tpu.CenteredRmsPropParameters.epsilon', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1129,
  serialized_end=1259,
)


_MDLADAGRADLIGHTPARAMETERS = _descriptor.Descriptor(
  name='MdlAdagradLightParameters',
  full_name='tensorflow.tpu.MdlAdagradLightParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='l2', full_name='tensorflow.tpu.MdlAdagradLightParameters.l2', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lr_power', full_name='tensorflow.tpu.MdlAdagradLightParameters.lr_power', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min_servable_mdl_benefit', full_name='tensorflow.tpu.MdlAdagradLightParameters.min_servable_mdl_benefit', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mdl_mix_in_margin', full_name='tensorflow.tpu.MdlAdagradLightParameters.mdl_mix_in_margin', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mdl_benefit_rampup_coeff', full_name='tensorflow.tpu.MdlAdagradLightParameters.mdl_benefit_rampup_coeff', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mdl_min_weight', full_name='tensorflow.tpu.MdlAdagradLightParameters.mdl_min_weight', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='benefit_revisit_scale', full_name='tensorflow.tpu.MdlAdagradLightParameters.benefit_revisit_scale', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_event_benefit', full_name='tensorflow.tpu.MdlAdagradLightParameters.max_event_benefit', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_total_benefit', full_name='tensorflow.tpu.MdlAdagradLightParameters.max_total_benefit', index=8,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mdl_hard_limit', full_name='tensorflow.tpu.MdlAdagradLightParameters.mdl_hard_limit', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hard_limit_min_benefit', full_name='tensorflow.tpu.MdlAdagradLightParameters.hard_limit_min_benefit', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mdl_regularize', full_name='tensorflow.tpu.MdlAdagradLightParameters.mdl_regularize', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1262,
  serialized_end=1675,
)


_ADADELTAPARAMETERS = _descriptor.Descriptor(
  name='AdadeltaParameters',
  full_name='tensorflow.tpu.AdadeltaParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rho', full_name='tensorflow.tpu.AdadeltaParameters.rho', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='epsilon', full_name='tensorflow.tpu.AdadeltaParameters.epsilon', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1677,
  serialized_end=1776,
)


_PROXIMALADAGRADPARAMETERS = _descriptor.Descriptor(
  name='ProximalAdagradParameters',
  full_name='tensorflow.tpu.ProximalAdagradParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='l1', full_name='tensorflow.tpu.ProximalAdagradParameters.l1', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='l2', full_name='tensorflow.tpu.ProximalAdagradParameters.l2', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1778,
  serialized_end=1856,
)


_ONLINEYOGIPARAMETERS = _descriptor.Descriptor(
  name='OnlineYogiParameters',
  full_name='tensorflow.tpu.OnlineYogiParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='l1', full_name='tensorflow.tpu.OnlineYogiParameters.l1', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='l2', full_name='tensorflow.tpu.OnlineYogiParameters.l2', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beta2', full_name='tensorflow.tpu.OnlineYogiParameters.beta2', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1858,
  serialized_end=1931,
)


_PROXIMALYOGIPARAMETERS = _descriptor.Descriptor(
  name='ProximalYogiParameters',
  full_name='tensorflow.tpu.ProximalYogiParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='l1', full_name='tensorflow.tpu.ProximalYogiParameters.l1', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='l2', full_name='tensorflow.tpu.ProximalYogiParameters.l2', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beta1', full_name='tensorflow.tpu.ProximalYogiParameters.beta1', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beta2', full_name='tensorflow.tpu.ProximalYogiParameters.beta2', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='epsilon', full_name='tensorflow.tpu.ProximalYogiParameters.epsilon', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1933,
  serialized_end=2040,
)


_FREQUENCYESTIMATORPARAMETERS = _descriptor.Descriptor(
  name='FrequencyEstimatorParameters',
  full_name='tensorflow.tpu.FrequencyEstimatorParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tau', full_name='tensorflow.tpu.FrequencyEstimatorParameters.tau', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_delta', full_name='tensorflow.tpu.FrequencyEstimatorParameters.max_delta', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='outlier_threshold', full_name='tensorflow.tpu.FrequencyEstimatorParameters.outlier_threshold', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_exponent', full_name='tensorflow.tpu.FrequencyEstimatorParameters.weight_exponent', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2042,
  serialized_end=2156,
)


_USERDEFINEDPROGRAMPARAMETERS = _descriptor.Descriptor(
  name='UserDefinedProgramParameters',
  full_name='tensorflow.tpu.UserDefinedProgramParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='program', full_name='tensorflow.tpu.UserDefinedProgramParameters.program', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='padding_values', full_name='tensorflow.tpu.UserDefinedProgramParameters.padding_values', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2158,
  serialized_end=2250,
)


_GRADIENTACCUMULATIONSTATUS = _descriptor.Descriptor(
  name='GradientAccumulationStatus',
  full_name='tensorflow.tpu.GradientAccumulationStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _GRADIENTACCUMULATIONSTATUS_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2252,
  serialized_end=2334,
)


_HOTIDREPLICATIONCONFIGURATION = _descriptor.Descriptor(
  name='HotIdReplicationConfiguration',
  full_name='tensorflow.tpu.HotIdReplicationConfiguration',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='tensorflow.tpu.HotIdReplicationConfiguration.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _HOTIDREPLICATIONCONFIGURATION_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2337,
  serialized_end=2492,
)


_OPTIMIZATIONPARAMETERS = _descriptor.Descriptor(
  name='OptimizationParameters',
  full_name='tensorflow.tpu.OptimizationParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='learning_rate', full_name='tensorflow.tpu.OptimizationParameters.learning_rate', index=0,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='clipping_limits', full_name='tensorflow.tpu.OptimizationParameters.clipping_limits', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gradient_clipping_limits', full_name='tensorflow.tpu.OptimizationParameters.gradient_clipping_limits', index=2,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_decay_factor', full_name='tensorflow.tpu.OptimizationParameters.weight_decay_factor', index=3,
      number=16, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multiply_weight_decay_factor_by_learning_rate', full_name='tensorflow.tpu.OptimizationParameters.multiply_weight_decay_factor_by_learning_rate', index=4,
      number=22, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gradient_accumulation_status', full_name='tensorflow.tpu.OptimizationParameters.gradient_accumulation_status', index=5,
      number=17, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hot_id_replication_configuration', full_name='tensorflow.tpu.OptimizationParameters.hot_id_replication_configuration', index=6,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adagrad', full_name='tensorflow.tpu.OptimizationParameters.adagrad', index=7,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bounded_adagrad', full_name='tensorflow.tpu.OptimizationParameters.bounded_adagrad', index=8,
      number=19, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stochastic_gradient_descent', full_name='tensorflow.tpu.OptimizationParameters.stochastic_gradient_descent', index=9,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ftrl', full_name='tensorflow.tpu.OptimizationParameters.ftrl', index=10,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adam', full_name='tensorflow.tpu.OptimizationParameters.adam', index=11,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='momentum', full_name='tensorflow.tpu.OptimizationParameters.momentum', index=12,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rms_prop', full_name='tensorflow.tpu.OptimizationParameters.rms_prop', index=13,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='centered_rms_prop', full_name='tensorflow.tpu.OptimizationParameters.centered_rms_prop', index=14,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mdl_adagrad_light', full_name='tensorflow.tpu.OptimizationParameters.mdl_adagrad_light', index=15,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adadelta', full_name='tensorflow.tpu.OptimizationParameters.adadelta', index=16,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proximal_adagrad', full_name='tensorflow.tpu.OptimizationParameters.proximal_adagrad', index=17,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='online_yogi', full_name='tensorflow.tpu.OptimizationParameters.online_yogi', index=18,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proximal_yogi', full_name='tensorflow.tpu.OptimizationParameters.proximal_yogi', index=19,
      number=21, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='frequency_estimator', full_name='tensorflow.tpu.OptimizationParameters.frequency_estimator', index=20,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user_defined_program', full_name='tensorflow.tpu.OptimizationParameters.user_defined_program', index=21,
      number=24, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='parameters', full_name='tensorflow.tpu.OptimizationParameters.parameters',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=2495,
  serialized_end=3957,
)


_STATEVARIABLESPECIFICATION_USERDEFINED = _descriptor.Descriptor(
  name='UserDefined',
  full_name='tensorflow.tpu.StateVariableSpecification.UserDefined',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='padding_initial_value', full_name='tensorflow.tpu.StateVariableSpecification.UserDefined.padding_initial_value', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4175,
  serialized_end=4219,
)

_STATEVARIABLESPECIFICATION_FILLWITHCONSTANT = _descriptor.Descriptor(
  name='FillWithConstant',
  full_name='tensorflow.tpu.StateVariableSpecification.FillWithConstant',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='initial_value', full_name='tensorflow.tpu.StateVariableSpecification.FillWithConstant.initial_value', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4221,
  serialized_end=4262,
)

_STATEVARIABLESPECIFICATION = _descriptor.Descriptor(
  name='StateVariableSpecification',
  full_name='tensorflow.tpu.StateVariableSpecification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='tensorflow.tpu.StateVariableSpecification.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user_defined', full_name='tensorflow.tpu.StateVariableSpecification.user_defined', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fill_with_constant', full_name='tensorflow.tpu.StateVariableSpecification.fill_with_constant', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_STATEVARIABLESPECIFICATION_USERDEFINED, _STATEVARIABLESPECIFICATION_FILLWITHCONSTANT, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='usage', full_name='tensorflow.tpu.StateVariableSpecification.usage',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=3960,
  serialized_end=4271,
)

_CLIPPINGLIMITS.fields_by_name['lower'].message_type = google_dot_protobuf_dot_wrappers__pb2._FLOATVALUE
_CLIPPINGLIMITS.fields_by_name['upper'].message_type = google_dot_protobuf_dot_wrappers__pb2._FLOATVALUE
_LEARNINGRATE.fields_by_name['dynamic'].message_type = _DYNAMICLEARNINGRATE
_LEARNINGRATE.oneofs_by_name['learning_rate'].fields.append(
  _LEARNINGRATE.fields_by_name['constant'])
_LEARNINGRATE.fields_by_name['constant'].containing_oneof = _LEARNINGRATE.oneofs_by_name['learning_rate']
_LEARNINGRATE.oneofs_by_name['learning_rate'].fields.append(
  _LEARNINGRATE.fields_by_name['dynamic'])
_LEARNINGRATE.fields_by_name['dynamic'].containing_oneof = _LEARNINGRATE.oneofs_by_name['learning_rate']
_USERDEFINEDPROGRAMPARAMETERS.fields_by_name['program'].message_type = tensorflow_dot_compiler_dot_xla_dot_service_dot_hlo__pb2._HLOMODULEPROTO
_GRADIENTACCUMULATIONSTATUS_STATUS.containing_type = _GRADIENTACCUMULATIONSTATUS
_HOTIDREPLICATIONCONFIGURATION.fields_by_name['status'].enum_type = _HOTIDREPLICATIONCONFIGURATION_STATUS
_HOTIDREPLICATIONCONFIGURATION_STATUS.containing_type = _HOTIDREPLICATIONCONFIGURATION
_OPTIMIZATIONPARAMETERS.fields_by_name['learning_rate'].message_type = _LEARNINGRATE
_OPTIMIZATIONPARAMETERS.fields_by_name['clipping_limits'].message_type = _CLIPPINGLIMITS
_OPTIMIZATIONPARAMETERS.fields_by_name['gradient_clipping_limits'].message_type = _CLIPPINGLIMITS
_OPTIMIZATIONPARAMETERS.fields_by_name['gradient_accumulation_status'].enum_type = _GRADIENTACCUMULATIONSTATUS_STATUS
_OPTIMIZATIONPARAMETERS.fields_by_name['hot_id_replication_configuration'].message_type = _HOTIDREPLICATIONCONFIGURATION
_OPTIMIZATIONPARAMETERS.fields_by_name['adagrad'].message_type = _ADAGRADPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['bounded_adagrad'].message_type = _BOUNDEDADAGRADPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['stochastic_gradient_descent'].message_type = _STOCHASTICGRADIENTDESCENTPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['ftrl'].message_type = _FTRLPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['adam'].message_type = _ADAMPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['momentum'].message_type = _MOMENTUMPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['rms_prop'].message_type = _RMSPROPPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['centered_rms_prop'].message_type = _CENTEREDRMSPROPPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['mdl_adagrad_light'].message_type = _MDLADAGRADLIGHTPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['adadelta'].message_type = _ADADELTAPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['proximal_adagrad'].message_type = _PROXIMALADAGRADPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['online_yogi'].message_type = _ONLINEYOGIPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['proximal_yogi'].message_type = _PROXIMALYOGIPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['frequency_estimator'].message_type = _FREQUENCYESTIMATORPARAMETERS
_OPTIMIZATIONPARAMETERS.fields_by_name['user_defined_program'].message_type = _USERDEFINEDPROGRAMPARAMETERS
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['adagrad'])
_OPTIMIZATIONPARAMETERS.fields_by_name['adagrad'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['bounded_adagrad'])
_OPTIMIZATIONPARAMETERS.fields_by_name['bounded_adagrad'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['stochastic_gradient_descent'])
_OPTIMIZATIONPARAMETERS.fields_by_name['stochastic_gradient_descent'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['ftrl'])
_OPTIMIZATIONPARAMETERS.fields_by_name['ftrl'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['adam'])
_OPTIMIZATIONPARAMETERS.fields_by_name['adam'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['momentum'])
_OPTIMIZATIONPARAMETERS.fields_by_name['momentum'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['rms_prop'])
_OPTIMIZATIONPARAMETERS.fields_by_name['rms_prop'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['centered_rms_prop'])
_OPTIMIZATIONPARAMETERS.fields_by_name['centered_rms_prop'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['mdl_adagrad_light'])
_OPTIMIZATIONPARAMETERS.fields_by_name['mdl_adagrad_light'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['adadelta'])
_OPTIMIZATIONPARAMETERS.fields_by_name['adadelta'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['proximal_adagrad'])
_OPTIMIZATIONPARAMETERS.fields_by_name['proximal_adagrad'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['online_yogi'])
_OPTIMIZATIONPARAMETERS.fields_by_name['online_yogi'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['proximal_yogi'])
_OPTIMIZATIONPARAMETERS.fields_by_name['proximal_yogi'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['frequency_estimator'])
_OPTIMIZATIONPARAMETERS.fields_by_name['frequency_estimator'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters'].fields.append(
  _OPTIMIZATIONPARAMETERS.fields_by_name['user_defined_program'])
_OPTIMIZATIONPARAMETERS.fields_by_name['user_defined_program'].containing_oneof = _OPTIMIZATIONPARAMETERS.oneofs_by_name['parameters']
_STATEVARIABLESPECIFICATION_USERDEFINED.containing_type = _STATEVARIABLESPECIFICATION
_STATEVARIABLESPECIFICATION_FILLWITHCONSTANT.containing_type = _STATEVARIABLESPECIFICATION
_STATEVARIABLESPECIFICATION.fields_by_name['user_defined'].message_type = _STATEVARIABLESPECIFICATION_USERDEFINED
_STATEVARIABLESPECIFICATION.fields_by_name['fill_with_constant'].message_type = _STATEVARIABLESPECIFICATION_FILLWITHCONSTANT
_STATEVARIABLESPECIFICATION.oneofs_by_name['usage'].fields.append(
  _STATEVARIABLESPECIFICATION.fields_by_name['user_defined'])
_STATEVARIABLESPECIFICATION.fields_by_name['user_defined'].containing_oneof = _STATEVARIABLESPECIFICATION.oneofs_by_name['usage']
_STATEVARIABLESPECIFICATION.oneofs_by_name['usage'].fields.append(
  _STATEVARIABLESPECIFICATION.fields_by_name['fill_with_constant'])
_STATEVARIABLESPECIFICATION.fields_by_name['fill_with_constant'].containing_oneof = _STATEVARIABLESPECIFICATION.oneofs_by_name['usage']
DESCRIPTOR.message_types_by_name['ClippingLimits'] = _CLIPPINGLIMITS
DESCRIPTOR.message_types_by_name['DynamicLearningRate'] = _DYNAMICLEARNINGRATE
DESCRIPTOR.message_types_by_name['LearningRate'] = _LEARNINGRATE
DESCRIPTOR.message_types_by_name['AdagradParameters'] = _ADAGRADPARAMETERS
DESCRIPTOR.message_types_by_name['BoundedAdagradParameters'] = _BOUNDEDADAGRADPARAMETERS
DESCRIPTOR.message_types_by_name['StochasticGradientDescentParameters'] = _STOCHASTICGRADIENTDESCENTPARAMETERS
DESCRIPTOR.message_types_by_name['FtrlParameters'] = _FTRLPARAMETERS
DESCRIPTOR.message_types_by_name['AdamParameters'] = _ADAMPARAMETERS
DESCRIPTOR.message_types_by_name['MomentumParameters'] = _MOMENTUMPARAMETERS
DESCRIPTOR.message_types_by_name['RmsPropParameters'] = _RMSPROPPARAMETERS
DESCRIPTOR.message_types_by_name['CenteredRmsPropParameters'] = _CENTEREDRMSPROPPARAMETERS
DESCRIPTOR.message_types_by_name['MdlAdagradLightParameters'] = _MDLADAGRADLIGHTPARAMETERS
DESCRIPTOR.message_types_by_name['AdadeltaParameters'] = _ADADELTAPARAMETERS
DESCRIPTOR.message_types_by_name['ProximalAdagradParameters'] = _PROXIMALADAGRADPARAMETERS
DESCRIPTOR.message_types_by_name['OnlineYogiParameters'] = _ONLINEYOGIPARAMETERS
DESCRIPTOR.message_types_by_name['ProximalYogiParameters'] = _PROXIMALYOGIPARAMETERS
DESCRIPTOR.message_types_by_name['FrequencyEstimatorParameters'] = _FREQUENCYESTIMATORPARAMETERS
DESCRIPTOR.message_types_by_name['UserDefinedProgramParameters'] = _USERDEFINEDPROGRAMPARAMETERS
DESCRIPTOR.message_types_by_name['GradientAccumulationStatus'] = _GRADIENTACCUMULATIONSTATUS
DESCRIPTOR.message_types_by_name['HotIdReplicationConfiguration'] = _HOTIDREPLICATIONCONFIGURATION
DESCRIPTOR.message_types_by_name['OptimizationParameters'] = _OPTIMIZATIONPARAMETERS
DESCRIPTOR.message_types_by_name['StateVariableSpecification'] = _STATEVARIABLESPECIFICATION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ClippingLimits = _reflection.GeneratedProtocolMessageType('ClippingLimits', (_message.Message,), {
  'DESCRIPTOR' : _CLIPPINGLIMITS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.ClippingLimits)
  })
_sym_db.RegisterMessage(ClippingLimits)

DynamicLearningRate = _reflection.GeneratedProtocolMessageType('DynamicLearningRate', (_message.Message,), {
  'DESCRIPTOR' : _DYNAMICLEARNINGRATE,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.DynamicLearningRate)
  })
_sym_db.RegisterMessage(DynamicLearningRate)

LearningRate = _reflection.GeneratedProtocolMessageType('LearningRate', (_message.Message,), {
  'DESCRIPTOR' : _LEARNINGRATE,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.LearningRate)
  })
_sym_db.RegisterMessage(LearningRate)

AdagradParameters = _reflection.GeneratedProtocolMessageType('AdagradParameters', (_message.Message,), {
  'DESCRIPTOR' : _ADAGRADPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.AdagradParameters)
  })
_sym_db.RegisterMessage(AdagradParameters)

BoundedAdagradParameters = _reflection.GeneratedProtocolMessageType('BoundedAdagradParameters', (_message.Message,), {
  'DESCRIPTOR' : _BOUNDEDADAGRADPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.BoundedAdagradParameters)
  })
_sym_db.RegisterMessage(BoundedAdagradParameters)

StochasticGradientDescentParameters = _reflection.GeneratedProtocolMessageType('StochasticGradientDescentParameters', (_message.Message,), {
  'DESCRIPTOR' : _STOCHASTICGRADIENTDESCENTPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.StochasticGradientDescentParameters)
  })
_sym_db.RegisterMessage(StochasticGradientDescentParameters)

FtrlParameters = _reflection.GeneratedProtocolMessageType('FtrlParameters', (_message.Message,), {
  'DESCRIPTOR' : _FTRLPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.FtrlParameters)
  })
_sym_db.RegisterMessage(FtrlParameters)

AdamParameters = _reflection.GeneratedProtocolMessageType('AdamParameters', (_message.Message,), {
  'DESCRIPTOR' : _ADAMPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.AdamParameters)
  })
_sym_db.RegisterMessage(AdamParameters)

MomentumParameters = _reflection.GeneratedProtocolMessageType('MomentumParameters', (_message.Message,), {
  'DESCRIPTOR' : _MOMENTUMPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.MomentumParameters)
  })
_sym_db.RegisterMessage(MomentumParameters)

RmsPropParameters = _reflection.GeneratedProtocolMessageType('RmsPropParameters', (_message.Message,), {
  'DESCRIPTOR' : _RMSPROPPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.RmsPropParameters)
  })
_sym_db.RegisterMessage(RmsPropParameters)

CenteredRmsPropParameters = _reflection.GeneratedProtocolMessageType('CenteredRmsPropParameters', (_message.Message,), {
  'DESCRIPTOR' : _CENTEREDRMSPROPPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.CenteredRmsPropParameters)
  })
_sym_db.RegisterMessage(CenteredRmsPropParameters)

MdlAdagradLightParameters = _reflection.GeneratedProtocolMessageType('MdlAdagradLightParameters', (_message.Message,), {
  'DESCRIPTOR' : _MDLADAGRADLIGHTPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.MdlAdagradLightParameters)
  })
_sym_db.RegisterMessage(MdlAdagradLightParameters)

AdadeltaParameters = _reflection.GeneratedProtocolMessageType('AdadeltaParameters', (_message.Message,), {
  'DESCRIPTOR' : _ADADELTAPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.AdadeltaParameters)
  })
_sym_db.RegisterMessage(AdadeltaParameters)

ProximalAdagradParameters = _reflection.GeneratedProtocolMessageType('ProximalAdagradParameters', (_message.Message,), {
  'DESCRIPTOR' : _PROXIMALADAGRADPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.ProximalAdagradParameters)
  })
_sym_db.RegisterMessage(ProximalAdagradParameters)

OnlineYogiParameters = _reflection.GeneratedProtocolMessageType('OnlineYogiParameters', (_message.Message,), {
  'DESCRIPTOR' : _ONLINEYOGIPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.OnlineYogiParameters)
  })
_sym_db.RegisterMessage(OnlineYogiParameters)

ProximalYogiParameters = _reflection.GeneratedProtocolMessageType('ProximalYogiParameters', (_message.Message,), {
  'DESCRIPTOR' : _PROXIMALYOGIPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.ProximalYogiParameters)
  })
_sym_db.RegisterMessage(ProximalYogiParameters)

FrequencyEstimatorParameters = _reflection.GeneratedProtocolMessageType('FrequencyEstimatorParameters', (_message.Message,), {
  'DESCRIPTOR' : _FREQUENCYESTIMATORPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.FrequencyEstimatorParameters)
  })
_sym_db.RegisterMessage(FrequencyEstimatorParameters)

UserDefinedProgramParameters = _reflection.GeneratedProtocolMessageType('UserDefinedProgramParameters', (_message.Message,), {
  'DESCRIPTOR' : _USERDEFINEDPROGRAMPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.UserDefinedProgramParameters)
  })
_sym_db.RegisterMessage(UserDefinedProgramParameters)

GradientAccumulationStatus = _reflection.GeneratedProtocolMessageType('GradientAccumulationStatus', (_message.Message,), {
  'DESCRIPTOR' : _GRADIENTACCUMULATIONSTATUS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.GradientAccumulationStatus)
  })
_sym_db.RegisterMessage(GradientAccumulationStatus)

HotIdReplicationConfiguration = _reflection.GeneratedProtocolMessageType('HotIdReplicationConfiguration', (_message.Message,), {
  'DESCRIPTOR' : _HOTIDREPLICATIONCONFIGURATION,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.HotIdReplicationConfiguration)
  })
_sym_db.RegisterMessage(HotIdReplicationConfiguration)

OptimizationParameters = _reflection.GeneratedProtocolMessageType('OptimizationParameters', (_message.Message,), {
  'DESCRIPTOR' : _OPTIMIZATIONPARAMETERS,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.OptimizationParameters)
  })
_sym_db.RegisterMessage(OptimizationParameters)

StateVariableSpecification = _reflection.GeneratedProtocolMessageType('StateVariableSpecification', (_message.Message,), {

  'UserDefined' : _reflection.GeneratedProtocolMessageType('UserDefined', (_message.Message,), {
    'DESCRIPTOR' : _STATEVARIABLESPECIFICATION_USERDEFINED,
    '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow.tpu.StateVariableSpecification.UserDefined)
    })
  ,

  'FillWithConstant' : _reflection.GeneratedProtocolMessageType('FillWithConstant', (_message.Message,), {
    'DESCRIPTOR' : _STATEVARIABLESPECIFICATION_FILLWITHCONSTANT,
    '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow.tpu.StateVariableSpecification.FillWithConstant)
    })
  ,
  'DESCRIPTOR' : _STATEVARIABLESPECIFICATION,
  '__module__' : 'tensorflow.core.protobuf.tpu.optimization_parameters_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tpu.StateVariableSpecification)
  })
_sym_db.RegisterMessage(StateVariableSpecification)
_sym_db.RegisterMessage(StateVariableSpecification.UserDefined)
_sym_db.RegisterMessage(StateVariableSpecification.FillWithConstant)


# @@protoc_insertion_point(module_scope)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utils for saved model benchmarks."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tempfile
import time

import tensorflow as tf

from tensorflow.python.platform import gfile
from tensorflow.python.platform import test


def save_and_load_benchmark(app):
  """Util for saved model benchmarks."""
  trials = 3

  model = app(weights=None)
  model_name = app.__name__

  tmp_dir = test.get_temp_dir()
  gfile.MakeDirs(tmp_dir)
  save_dir = tempfile.mkdtemp(dir=tmp_dir)

  total_save_time = 0
  total_load_time = 0

  # Run one untimed iteration of saving/loading.
  model.save(save_dir, save_format='tf')
  tf.keras.models.load_model(save_dir)

  for _ in range(trials):
    start_time = time.time()
    model.save(save_dir, save_format='tf')
    total_save_time += time.time() - start_time

    start_time = time.time()
    tf.keras.models.load_model(save_dir)
    total_load_time += time.time() - start_time

  save_result = {
      'iters': trials,
      'wall_time': total_save_time / trials,
      'name': '{}.save'.format(model_name)
  }

  load_result = {
      'iters': trials,
      'wall_time': total_load_time / trials,
      'name': '{}.load'.format(model_name)
  }
  gfile.DeleteRecursively(save_dir)
  return save_result, load_result




# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Registrations for LinearOperator.matmul."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops.linalg import linear_operator
from tensorflow.python.ops.linalg import linear_operator_algebra
from tensorflow.python.ops.linalg import linear_operator_circulant
from tensorflow.python.ops.linalg import linear_operator_composition
from tensorflow.python.ops.linalg import linear_operator_diag
from tensorflow.python.ops.linalg import linear_operator_identity
from tensorflow.python.ops.linalg import linear_operator_lower_triangular
from tensorflow.python.ops.linalg import linear_operator_zeros
from tensorflow.python.ops.linalg import registrations_util


# By default, use a LinearOperatorComposition to delay the computation.
@linear_operator_algebra.RegisterMatmul(
    linear_operator.LinearOperator, linear_operator.LinearOperator)
def _matmul_linear_operator(linop_a, linop_b):
  """Generic matmul of two `LinearOperator`s."""
  is_square = registrations_util.is_square(linop_a, linop_b)
  is_non_singular = None
  is_self_adjoint = None
  is_positive_definite = None

  if is_square:
    is_non_singular = registrations_util.combined_non_singular_hint(
        linop_a, linop_b)
  elif is_square is False:  # pylint:disable=g-bool-id-comparison
    is_non_singular = False
    is_self_adjoint = False
    is_positive_definite = False

  return linear_operator_composition.LinearOperatorComposition(
      operators=[linop_a, linop_b],
      is_non_singular=is_non_singular,
      is_self_adjoint=is_self_adjoint,
      is_positive_definite=is_positive_definite,
      is_square=is_square,
  )

# Identity


@linear_operator_algebra.RegisterMatmul(
    linear_operator_identity.LinearOperatorIdentity,
    linear_operator.LinearOperator)
def _matmul_linear_operator_identity_left(identity, linop):
  del identity
  return linop


@linear_operator_algebra.RegisterMatmul(
    linear_operator.LinearOperator,
    linear_operator_identity.LinearOperatorIdentity)
def _matmul_linear_operator_identity_right(linop, identity):
  del identity
  return linop


@linear_operator_algebra.RegisterMatmul(
    linear_operator_identity.LinearOperatorScaledIdentity,
    linear_operator_identity.LinearOperatorScaledIdentity)
def _matmul_linear_operator_scaled_identity(linop_a, linop_b):
  """Matmul of two ScaledIdentity `LinearOperators`."""
  return linear_operator_identity.LinearOperatorScaledIdentity(
      num_rows=linop_a.domain_dimension_tensor(),
      multiplier=linop_a.multiplier * linop_b.multiplier,
      is_non_singular=registrations_util.combined_non_singular_hint(
          linop_a, linop_b),
      is_self_adjoint=registrations_util.combined_commuting_self_adjoint_hint(
          linop_a, linop_b),
      is_positive_definite=(
          registrations_util.combined_commuting_positive_definite_hint(
              linop_a, linop_b)),
      is_square=True)


# Zeros


@linear_operator_algebra.RegisterMatmul(
    linear_operator.LinearOperator,
    linear_operator_zeros.LinearOperatorZeros)
def _matmul_linear_operator_zeros_right(linop, zeros):
  if not zeros.is_square or not linop.is_square:
    raise ValueError("Matmul with non-square `LinearOperator`s or non-square "
                     "`LinearOperatorZeros` not supported at this time.")
  return zeros


@linear_operator_algebra.RegisterMatmul(
    linear_operator_zeros.LinearOperatorZeros,
    linear_operator.LinearOperator)
def _matmul_linear_operator_zeros_left(zeros, linop):
  if not zeros.is_square or not linop.is_square:
    raise ValueError("Matmul with non-square `LinearOperator`s or non-square "
                     "`LinearOperatorZeros` not supported at this time.")
  return zeros


# Diag.


@linear_operator_algebra.RegisterMatmul(
    linear_operator_diag.LinearOperatorDiag,
    linear_operator_diag.LinearOperatorDiag)
def _matmul_linear_operator_diag(linop_a, linop_b):
  return linear_operator_diag.LinearOperatorDiag(
      diag=linop_a.diag * linop_b.diag,
      is_non_singular=registrations_util.combined_non_singular_hint(
          linop_a, linop_b),
      is_self_adjoint=registrations_util.combined_commuting_self_adjoint_hint(
          linop_a, linop_b),
      is_positive_definite=(
          registrations_util.combined_commuting_positive_definite_hint(
              linop_a, linop_b)),
      is_square=True)


@linear_operator_algebra.RegisterMatmul(
    linear_operator_diag.LinearOperatorDiag,
    linear_operator_identity.LinearOperatorScaledIdentity)
def _matmul_linear_operator_diag_scaled_identity_right(
    linop_diag, linop_scaled_identity):
  return linear_operator_diag.LinearOperatorDiag(
      diag=linop_diag.diag * linop_scaled_identity.multiplier,
      is_non_singular=registrations_util.combined_non_singular_hint(
          linop_diag, linop_scaled_identity),
      is_self_adjoint=registrations_util.combined_commuting_self_adjoint_hint(
          linop_diag, linop_scaled_identity),
      is_positive_definite=(
          registrations_util.combined_commuting_positive_definite_hint(
              linop_diag, linop_scaled_identity)),
      is_square=True)


@linear_operator_algebra.RegisterMatmul(
    linear_operator_identity.LinearOperatorScaledIdentity,
    linear_operator_diag.LinearOperatorDiag)
def _matmul_linear_operator_diag_scaled_identity_left(
    linop_scaled_identity, linop_diag):
  return linear_operator_diag.LinearOperatorDiag(
      diag=linop_diag.diag * linop_scaled_identity.multiplier,
      is_non_singular=registrations_util.combined_non_singular_hint(
          linop_diag, linop_scaled_identity),
      is_self_adjoint=registrations_util.combined_commuting_self_adjoint_hint(
          linop_diag, linop_scaled_identity),
      is_positive_definite=(
          registrations_util.combined_commuting_positive_definite_hint(
              linop_diag, linop_scaled_identity)),
      is_square=True)


@linear_operator_algebra.RegisterMatmul(
    linear_operator_diag.LinearOperatorDiag,
    linear_operator_lower_triangular.LinearOperatorLowerTriangular)
def _matmul_linear_operator_diag_tril(linop_diag, linop_triangular):
  return linear_operator_lower_triangular.LinearOperatorLowerTriangular(
      tril=linop_diag.diag[..., None] * linop_triangular.to_dense(),
      is_non_singular=registrations_util.combined_non_singular_hint(
          linop_diag, linop_triangular),
      # This is safe to do since the Triangular matrix is only self-adjoint
      # when it is a diagonal matrix, and hence commutes.
      is_self_adjoint=registrations_util.combined_commuting_self_adjoint_hint(
          linop_diag, linop_triangular),
      is_positive_definite=None,
      is_square=True)


@linear_operator_algebra.RegisterMatmul(
    linear_operator_lower_triangular.LinearOperatorLowerTriangular,
    linear_operator_diag.LinearOperatorDiag)
def _matmul_linear_operator_tril_diag(linop_triangular, linop_diag):
  return linear_operator_lower_triangular.LinearOperatorLowerTriangular(
      tril=linop_triangular.to_dense() * linop_diag.diag,
      is_non_singular=registrations_util.combined_non_singular_hint(
          linop_diag, linop_triangular),
      # This is safe to do since the Triangular matrix is only self-adjoint
      # when it is a diagonal matrix, and hence commutes.
      is_self_adjoint=registrations_util.combined_commuting_self_adjoint_hint(
          linop_diag, linop_triangular),
      is_positive_definite=None,
      is_square=True)

# Circulant.


@linear_operator_algebra.RegisterMatmul(
    linear_operator_circulant.LinearOperatorCirculant,
    linear_operator_circulant.LinearOperatorCirculant)
def _matmul_linear_operator_circulant_circulant(linop_a, linop_b):
  return linear_operator_circulant.LinearOperatorCirculant(
      spectrum=linop_a.spectrum * linop_b.spectrum,
      is_non_singular=registrations_util.combined_non_singular_hint(
          linop_a, linop_b),
      is_self_adjoint=registrations_util.combined_commuting_self_adjoint_hint(
          linop_a, linop_b),
      is_positive_definite=(
          registrations_util.combined_commuting_positive_definite_hint(
              linop_a, linop_b)),
      is_square=True)



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Class to hold a library of OpDefs and use it to create Brain operations."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from google.protobuf import text_format
from tensorflow.core.framework import attr_value_pb2
from tensorflow.core.framework import tensor_pb2
from tensorflow.core.framework import tensor_shape_pb2
from tensorflow.core.framework import types_pb2
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import op_callbacks
from tensorflow.python.framework import op_def_registry
from tensorflow.python.framework import ops
from tensorflow.python.framework import tensor_shape
from tensorflow.python.platform import tf_logging as logging
from tensorflow.python.util import _pywrap_utils
from tensorflow.python.util import compat
from tensorflow.python.util import tf_contextlib


def _Attr(op_def, name):
  for attr in op_def.attr:
    if attr.name == name:
      return attr
  raise TypeError("Inconsistent OpDef for '%s', missing attr '%s'" %
                  (op_def.name, name))


def _AttrValue(attr_protos, name):
  if name in attr_protos:
    return attr_protos[name]
  raise TypeError("Inconsistent OpDef, missing attr '%s' from '%s'." %
                  (name, attr_protos))


def _SatisfiesTypeConstraint(dtype, attr_def, param_name):
  if attr_def.HasField("allowed_values"):
    allowed_list = attr_def.allowed_values.list.type
    if dtype not in allowed_list:
      raise TypeError(
          "Value passed to parameter '%s' has DataType %s not in list of "
          "allowed values: %s" %
          (param_name, dtypes.as_dtype(dtype).name,
           ", ".join(dtypes.as_dtype(x).name for x in allowed_list)))


def _SatisfiesLengthConstraint(length, attr_def, param_name, op_type_name):
  if attr_def.has_minimum and length < attr_def.minimum:
    raise ValueError("Attr '%s' of '%s' Op passed list of length %d "
                     "less than minimum %d." %
                     (param_name, op_type_name, length, attr_def.minimum))


def _SatisfiesAllowedStringsConstraint(value, attr_def, arg_name, op_type_name):
  if value not in attr_def.allowed_values.list.s:
    raise ValueError(
        "Attr '%s' of '%s' Op passed string '%s' not in: \"%s\"." %
        (arg_name, op_type_name, compat.as_text(value), '", "'.join(
            map(compat.as_text, attr_def.allowed_values.list.s))))


def _SatisfiesIntMinimumConstraint(value, attr_def, arg_name, op_type_name):
  if value < attr_def.minimum:
    raise ValueError("Attr '%s' of '%s' Op passed %d less than minimum %d." %
                     (arg_name, op_type_name, value, attr_def.minimum))


def _IsListParameter(arg):
  if arg.number_attr:
    return True
  elif arg.type_list_attr:
    return True
  return False


def _NumTypeFields(arg):
  num = 0
  if arg.type != types_pb2.DT_INVALID: num += 1
  if arg.type_attr: num += 1
  if arg.type_list_attr: num += 1
  return num


def _IsListValue(v):
  return isinstance(v, (list, tuple))


def _Flatten(l):
  """Converts [1, 2, [3, 4], [5]] to [1, 2, 3, 4, 5]."""
  # [1, 2, [3, 4], [5]] -> [[1], [2], [3, 4], [5]]
  l_of_l = [x if _IsListValue(x) else [x] for x in l]
  # [[1], [2], [3, 4], [5]] -> [1, 2, 3, 4, 5]
  return [item for sublist in l_of_l for item in sublist]


def _Restructure(l, structure):
  """Returns the elements of list l structured according to the given structure.

  A structure is represented by a list whose elements are either
  `None` or a non-negative integer. `None` corresponds to a single
  element in the output list, and an integer N corresponds to a nested
  list of length N.

  The function returns a data structure whose shape is given by
  `structure`, and whose elements are taken from `l`. If `structure`
  is a singleton, the function returns the single data structure
  implied by the 0th element of `structure`. For example:

      _Restructure(["foo", "bar", "baz", "qux"], [None, 2, None])
        -> ["foo", ["bar", "baz"], "qux"]

      _Restructure(["foo"], [None]) -> "foo"

      _Restructure(["foo"], [1]) -> ["foo"]

      _Restructure([], [0]) -> []

  Args:
    l: A list.
    structure: A list whose elements are either `None` or a non-negative
      integer.

  Returns:
    The elements of `l`, restructured according to `structure`. If
    `structure` is a list of length 1, this function returns the
    single data structure implied by `structure[0]`.

  """
  result = []
  current_index = 0
  for element in structure:
    if element is None:
      result.append(l[current_index])
      current_index += 1
    else:
      result.append(l[current_index:current_index+element])
      current_index += element

  if len(result) == 1:
    return result[0]
  else:
    return tuple(result)


def _MakeFloat(v, arg_name):
  if not isinstance(v, compat.real_types):
    raise TypeError("Expected float for argument '%s' not %s." %
                    (arg_name, repr(v)))
  return float(v)


def _MakeInt(v, arg_name):
  if isinstance(v, six.string_types):
    raise TypeError("Expected int for argument '%s' not %s." %
                    (arg_name, repr(v)))
  try:
    return int(v)
  except (ValueError, TypeError):
    raise TypeError("Expected int for argument '%s' not %s." %
                    (arg_name, repr(v)))


def _MakeStr(v, arg_name):
  if not isinstance(v, compat.bytes_or_text_types):
    raise TypeError("Expected string for argument '%s' not %s." %
                    (arg_name, repr(v)))
  return compat.as_bytes(v)  # Convert unicode strings to bytes.


def _MakeBool(v, arg_name):
  if not isinstance(v, bool):
    raise TypeError("Expected bool for argument '%s' not %s." %
                    (arg_name, repr(v)))
  return v


def _MakeType(v, arg_name):
  try:
    v = dtypes.as_dtype(v).base_dtype
  except TypeError:
    raise TypeError("Expected DataType for argument '%s' not %s." %
                    (arg_name, repr(v)))
  return v.as_datatype_enum


def _MakeShape(v, arg_name):
  """Convert v into a TensorShapeProto."""
  # Args:
  #   v: A TensorShapeProto, a list of ints, or a tensor_shape.TensorShape.
  #   arg_name: String, for error messages.

  # Returns:
  #   A TensorShapeProto.
  if isinstance(v, tensor_shape_pb2.TensorShapeProto):
    for d in v.dim:
      if d.name:
        logging.warning("Warning: TensorShapeProto with a named dimension: %s",
                        str(v))
        break
    return v
  try:
    return tensor_shape.as_shape(v).as_proto()
  except TypeError as e:
    raise TypeError("Error converting %s to a TensorShape: %s" % (arg_name, e))
  except ValueError as e:
    raise ValueError("Error converting %s to a TensorShape: %s" % (arg_name, e))


def _MakeTensor(v, arg_name):
  """Ensure v is a TensorProto."""
  if isinstance(v, tensor_pb2.TensorProto):
    return v
  raise TypeError(
      "Don't know how to convert %s to a TensorProto for argument '%s'" %
      (repr(v), arg_name))


def _MakeFunc(v, arg_name):
  """Ensure v is a func."""
  if isinstance(v, attr_value_pb2.NameAttrList):
    return v
  if isinstance(v, compat.bytes_or_text_types):
    fn_attr = attr_value_pb2.NameAttrList(name=v)
  elif hasattr(v, "add_to_graph"):
    v.add_to_graph(ops.get_default_graph())
    if hasattr(v, "_as_name_attr_list"):
      fn_attr = v._as_name_attr_list  # pylint: disable=protected-access
    else:
      fn_attr = attr_value_pb2.NameAttrList(name=v.name)
  else:
    raise TypeError("Don't know how to convert {} to a func for "
                    "argument {}".format(v, arg_name))
  return fn_attr


# pylint: disable=g-doc-return-or-yield
@tf_contextlib.contextmanager
def _MaybeColocateWith(inputs):
  """A context manager for (maybe) colocating with a list of input tensors.

  Args:
    inputs: A list of `Tensor` or `Operation` objects.

  Returns:
    A context manager.
  """
  if not inputs:
    yield
  else:
    # NOTE(mrry): The `ops.colocate_with()` function accepts only a single
    # op or tensor, so we create one context manager per element in the list.
    with ops.colocate_with(inputs[0]), _MaybeColocateWith(inputs[1:]):
      yield
# pylint: enable=g-doc-return-or-yield


def apply_op(op_type_name, name=None, **keywords):  # pylint: disable=invalid-name
  """Add a node invoking a registered Op to a graph.

  Example usage:
     # input1 and input2 can be Tensors or anything ops.convert_to_tensor()
     # will convert to a Tensor.
     op_def_library.apply_op("op", input1=input1, input2=input2)
     # Can specify a node name.
     op_def_library.apply_op("op", input1=input1, name="node_name")
     # Must use keyword arguments, with the names specified in the OpDef.
     op_def_library.apply_op("op", input_name=input, attr_name=attr)

  All attrs must either be inferred from an input or specified.
  (If inferred, the attr must not be specified.)  If an attr has a default
  value specified in the Op's OpDef, then you may pass None as the value
  of that attr to get the default.

  Args:
    op_type_name: string. Must match the name field of a registered Op.
    name: string. Optional name of the created op.
    **keywords: input Tensor and attr arguments specified by name,
      and optional parameters to pass when constructing the Operation.

  Returns:
    The Tensor(s) representing the output of the operation, or the Operation
    itself if there are no outputs.

  Raises:
    RuntimeError: On some errors.
    TypeError: On some errors.
    ValueError: On some errors.
  """
  output_structure, is_stateful, op, outputs = _apply_op_helper(
      op_type_name, name, **keywords)
  if output_structure:
    res = _Restructure(ops.convert_n_to_tensor(outputs), output_structure)
    if isinstance(res, list) and not res and is_stateful:
      return op
    else:
      return res
  else:
    return op


def _apply_op_helper(op_type_name, name=None, **keywords):  # pylint: disable=invalid-name
  """Implementation of apply_op that returns output_structure, op."""
  op_def = op_def_registry.get(op_type_name)
  if op_def is None:
    raise RuntimeError("Unrecognized Op name " + op_type_name)

  # Determine the graph context.
  try:
    # Need to flatten all the arguments into a list.
    # pylint: disable=protected-access
    g = ops._get_graph_from_inputs(_Flatten(keywords.values()))
    # pylint: enable=protected-access
  except AssertionError as e:
    raise RuntimeError(
        "Cannot determine graph for Op '%s' due to: %s"
        % (op_type_name, e.message))

  # Default name if not specified.
  if name is None:
    name = op_type_name

  # Check for deprecation
  deprecation_version = op_def.deprecation.version
  if deprecation_version:
    producer = g.graph_def_versions.producer
    if producer >= deprecation_version:
      raise NotImplementedError(
          ("Op %s is not available in GraphDef version %d. "
           "It has been removed in version %d. %s.") %
          (op_type_name, producer, deprecation_version,
           op_def.deprecation.explanation))

  # Fill in the list of default types for all "type" attrs.  This
  # will be used to choose a preferred dtype to convert to in the
  # absence of input type information.
  #
  # TODO(b/31302892): Currently the defaults don't work in the right
  # way if you have two inputs, one of whose type resolution depends
  # on the other.  Handling this will require restructuring this code
  # significantly.
  default_type_attr_map = {}
  allowed_list_attr_map = {}
  for attr_def in op_def.attr:
    if attr_def.type != "type":
      continue
    key = attr_def.name
    if attr_def.HasField("default_value"):
      default_type_attr_map[key] = dtypes.as_dtype(
          attr_def.default_value.type)
    if attr_def.HasField("allowed_values"):
      allowed_list_attr_map[key] = attr_def.allowed_values.list.type

  # Requires that op_def has passed validation (using the C++
  # ValidateOpDef() from ../framework/op_def_util.h).
  attrs = {}
  inputs = []
  input_types = []
  with g.as_default(), ops.name_scope(name) as scope:

    # Perform input type inference
    inferred_from = {}
    for input_arg in op_def.input_arg:
      input_name = input_arg.name
      if input_name in keywords:
        values = keywords.pop(input_name)
      elif input_name + "_" in keywords:
        # Handle the case where the name is a keyword or built-in
        # for Python so we use the name + _ instead.
        input_name += "_"
        values = keywords.pop(input_name)
      else:
        raise TypeError("No argument for input " + input_name)

      # Goals:
      # * Convert values to Tensors if it contains constants.
      # * Verify that values is a list if that matches the input_arg's
      #   type.
      # * If the input_arg's type is determined by attrs, either set
      #   those attrs and validate those attr values are legal (if
      #   they have not yet been set) or validate the input matches
      #   the type indicated by the attrs (if they have already been
      #   inferred via an earlier input).
      # * If the input_arg has an explicit type, make sure the input
      #   conforms.

      if _IsListParameter(input_arg):
        if not _IsListValue(values):
          raise TypeError(
              "Expected list for '%s' argument to '%s' Op, not %s." %
              (input_name, op_type_name, values))
        # In cases where we expect all elements of the list to have the
        # same dtype, try to cast non-Tensor elements to that type.
        dtype = None
        default_dtype = None
        if input_arg.type != types_pb2.DT_INVALID:
          dtype = input_arg.type
        elif input_arg.number_attr:
          if input_arg.type_attr in attrs:
            dtype = attrs[input_arg.type_attr]
          else:
            for t in values:
              if isinstance(t, ops.Tensor):
                dtype = t.dtype
                break

          # dtype still not found, prefer using the default dtype
          # from the attr.
          if dtype is None and input_arg.type_attr in default_type_attr_map:
            default_dtype = default_type_attr_map[input_arg.type_attr]

        try:
          if not input_arg.is_ref and dtype:
            dtype = dtypes.as_dtype(dtype).base_dtype
          values = ops.internal_convert_n_to_tensor(
              values,
              name=input_arg.name,
              dtype=dtype if dtype else None,
              preferred_dtype=default_dtype,
              as_ref=input_arg.is_ref)
          if input_arg.number_attr and len(
              set(v.dtype.base_dtype for v in values)) > 1:
            raise TypeError()  # All types should match.
        except (TypeError, ValueError):
          # What types does the conversion function think values have?
          observed_types = []
          for value in values:
            try:
              converted_value = ops.convert_to_tensor(
                  value, as_ref=input_arg.is_ref)
              observed_types.append(converted_value.dtype.base_dtype.name)
            except (TypeError, ValueError):
              observed_types.append("<NOT CONVERTIBLE TO TENSOR>")
          observed = ", ".join(observed_types)

          prefix = (
              "Tensors in list passed to '%s' of '%s' Op have types [%s]" %
              (input_name, op_type_name, observed))
          if input_arg.number_attr:
            if input_arg.type != types_pb2.DT_INVALID:
              raise TypeError("%s that do not match expected type %s." %
                              (prefix, dtype.name))
            elif input_arg.type_attr in attrs:
              raise TypeError("%s that do not match type %s inferred from "
                              "earlier arguments." %
                              (prefix, dtype.name))
            else:
              raise TypeError("%s that don't all match." % prefix)
          else:
            raise TypeError(
                "%s that are invalid. Tensors: %s" % (prefix, values))

        types = [x.dtype for x in values]
        inputs.extend(values)
      else:
        # In cases where we have an expected type, try to convert non-Tensor
        # arguments to that type.
        dtype = None
        default_dtype = None
        allowed_list = None
        if input_arg.type != types_pb2.DT_INVALID:
          dtype = input_arg.type
        elif input_arg.type_attr in attrs:
          dtype = attrs[input_arg.type_attr]
        elif input_arg.type_attr in default_type_attr_map:
          # The dtype could not be inferred solely from the inputs,
          # so we prefer the attr's default, so code that adds a new attr
          # with a default is backwards compatible.
          default_dtype = default_type_attr_map[input_arg.type_attr]
          allowed_list = allowed_list_attr_map.get(input_arg.type_attr)

        try:
          # First see if we can get a valid dtype with the default conversion
          # and see if it matches an allowed dtypes. Some ops like ConcatV2 may
          # not list allowed dtypes, in which case we should skip this.
          if dtype is None and allowed_list:
            inferred = None
            try:
              inferred = ops.convert_to_tensor(
                  values, name=input_arg.name, as_ref=input_arg.is_ref)
            except TypeError as err:
              # When converting a python object such as a list of Dimensions, we
              # need a dtype to be specified, thus tensor conversion may throw
              # an exception which we will ignore and try again below.
              pass

            # If we did not match an allowed dtype, try again with the default
            # dtype. This could be because we have an empty tensor and thus we
            # picked the wrong type.
            if inferred is not None and inferred.dtype in allowed_list:
              values = inferred
            else:
              values = ops.convert_to_tensor(
                  values,
                  name=input_arg.name,
                  as_ref=input_arg.is_ref,
                  preferred_dtype=default_dtype)
          else:
            values = ops.convert_to_tensor(
                values,
                name=input_arg.name,
                dtype=dtype,
                as_ref=input_arg.is_ref,
                preferred_dtype=default_dtype)
        except TypeError as err:
          if dtype is None:
            raise err
          else:
            raise TypeError(
                "Expected %s passed to parameter '%s' of op '%s', got %s of "
                "type '%s' instead. Error: %s" %
                (dtypes.as_dtype(dtype).name, input_arg.name, op_type_name,
                 repr(values), type(values).__name__, err))
        except ValueError:
          # What type does convert_to_tensor think it has?
          try:
            observed = ops.convert_to_tensor(
                values, as_ref=input_arg.is_ref).dtype.name
          except ValueError as err:
            raise ValueError(
                "Tried to convert '%s' to a tensor and failed. Error: %s" %
                (input_name, err))
          prefix = ("Input '%s' of '%s' Op has type %s that does not match" %
                    (input_name, op_type_name, observed))
          if input_arg.type != types_pb2.DT_INVALID:
            raise TypeError("%s expected type of %s." %
                            (prefix, dtypes.as_dtype(input_arg.type).name))
          else:
            # Update the maps with the default, if needed.
            k = input_arg.type_attr
            if k in default_type_attr_map:
              if k not in attrs:
                attrs[k] = default_type_attr_map[k]
                if k not in inferred_from:
                  inferred_from[k] = "Default in OpDef"

            raise TypeError(
                "%s type %s of argument '%s'." %
                (prefix, dtypes.as_dtype(attrs[input_arg.type_attr]).name,
                 inferred_from[input_arg.type_attr]))

        types = [values.dtype]
        inputs.append(values)
      base_types = [x.base_dtype for x in types]

      if input_arg.number_attr:
        # <number-attr> * <type> or <number-attr> * <type-attr>
        if input_arg.number_attr in attrs:
          if len(values) != attrs[input_arg.number_attr]:
            raise ValueError(
                "List argument '%s' to '%s' Op with length %d must match "
                "length %d of argument '%s'." %
                (input_name, op_type_name, len(values),
                 attrs[input_arg.number_attr],
                 inferred_from[input_arg.number_attr]))
        else:
          attrs[input_arg.number_attr] = len(values)
          inferred_from[input_arg.number_attr] = input_name
          num_attr = _Attr(op_def, input_arg.number_attr)
          if num_attr.has_minimum and len(values) < num_attr.minimum:
            raise ValueError(
                "List argument '%s' to '%s' Op with length %d shorter "
                "than minimum length %d." %
                (input_name, op_type_name, len(values), num_attr.minimum))
        # All tensors must have the same base type.
        if any(bt != base_types[0] for bt in base_types):
          raise TypeError(
              "All tensors passed to '%s' of '%s' Op "
              "must have the same type." %
              (input_name, op_type_name))
        if input_arg.type != types_pb2.DT_INVALID:
          # <number-attr> * <type> case
          if base_types and base_types[0] != input_arg.type:
            assert False, "Unreachable"
        elif input_arg.type_attr in attrs:
          # <number-attr> * <type-attr> case, where <type-attr> already
          # has an inferred value.
          if base_types and base_types[0] != attrs[input_arg.type_attr]:
            assert False, "Unreachable"
        else:
          # <number-attr> * <type-attr> case, where we are now setting
          # the <type-attr> based on this input
          if not base_types:
            # If it's in default_type_attr_map, then wait to set it
            # (in "process remaining attrs", below).
            if input_arg.type_attr not in default_type_attr_map:
              raise TypeError(
                  "Don't know how to infer type variable from empty input "
                  "list passed to input '%s' of '%s' Op." %
                  (input_name, op_type_name))
          else:
            attrs[input_arg.type_attr] = base_types[0]
            inferred_from[input_arg.type_attr] = input_name
            type_attr = _Attr(op_def, input_arg.type_attr)
            _SatisfiesTypeConstraint(base_types[0], type_attr,
                                     param_name=input_name)
      elif input_arg.type_attr:
        # <type-attr>
        attr_value = base_types[0]
        if input_arg.type_attr in attrs:
          if attrs[input_arg.type_attr] != attr_value:
            raise TypeError(
                "Input '%s' of '%s' Op has type %s that does not "
                "match type %s of argument '%s'." %
                (input_name, op_type_name, dtypes.as_dtype(attr_value).name,
                 dtypes.as_dtype(attrs[input_arg.type_attr]).name,
                 inferred_from[input_arg.type_attr]))
        else:
          for base_type in base_types:
            _SatisfiesTypeConstraint(base_type,
                                     _Attr(op_def, input_arg.type_attr),
                                     param_name=input_name)
          attrs[input_arg.type_attr] = attr_value
          inferred_from[input_arg.type_attr] = input_name
      elif input_arg.type_list_attr:
        # <type-list-attr>
        attr_value = base_types
        if input_arg.type_list_attr in attrs:
          if attrs[input_arg.type_list_attr] != attr_value:
            raise TypeError(
                "Input '%s' of '%s' Op has type list of %s that does not "
                "match type list %s of argument '%s'." %
                (input_name, op_type_name,
                 ", ".join(dtypes.as_dtype(x).name for x in attr_value),
                 ", ".join(dtypes.as_dtype(x).name
                           for x in attrs[input_arg.type_list_attr]),
                 inferred_from[input_arg.type_list_attr]))
        else:
          for base_type in base_types:
            _SatisfiesTypeConstraint(base_type,
                                     _Attr(op_def, input_arg.type_list_attr),
                                     param_name=input_name)
          attrs[input_arg.type_list_attr] = attr_value
          inferred_from[input_arg.type_list_attr] = input_name
      else:
        # single Tensor with specified type
        if base_types[0] != input_arg.type:
          assert False, "Unreachable"

      if input_arg.is_ref:
        if not all(x._is_ref_dtype for x in types):  # pylint: disable=protected-access
          raise TypeError(
              ("'%s' Op requires that input '%s' be a mutable tensor "
               "(e.g.: a tf.Variable)") % (op_type_name, input_name))
        input_types.extend(types)
      else:
        input_types.extend(base_types)

    # Process remaining attrs
    for attr in op_def.attr:
      # Skip attrs that have already had their values inferred
      if attr.name in attrs:
        if attr.name in keywords:
          raise TypeError(
              "Should not specify value for inferred attr '%s'." % attr.name)
        continue
      if attr.name in keywords:
        attrs[attr.name] = keywords.pop(attr.name)
      elif attr.name + "_" in keywords:
        # Attrs whose names match Python keywords have an extra '_'
        # appended, so we must check for that as well.
        attrs[attr.name] = keywords.pop(attr.name + "_")
      elif attr.name in default_type_attr_map:
        attrs[attr.name] = default_type_attr_map[attr.name]
        inferred_from.setdefault(attr.name, "Default in OpDef")
      else:
        raise TypeError("No argument for attr " + attr.name)

    # Convert attr values to AttrValue protos.
    attr_protos = {}
    for attr_def in op_def.attr:
      key = attr_def.name
      value = attrs[key]

      if attr_def.HasField("default_value") and value is None:
        attr_value = attr_value_pb2.AttrValue()
        attr_value.CopyFrom(attr_def.default_value)
        attr_protos[key] = attr_value
        continue

      attr_value = value_to_attr_value(value, attr_def.type, key)
      if attr_def.type.startswith("list("):
        _SatisfiesLengthConstraint(len(value), attr_def, key, op_type_name)
      if attr_def.HasField("allowed_values"):
        if attr_def.type == "string":
          _SatisfiesAllowedStringsConstraint(attr_value.s, attr_def, key,
                                             op_type_name)
        elif attr_def.type == "list(string)":
          for value in attr_value.list.s:
            _SatisfiesAllowedStringsConstraint(value, attr_def, key,
                                               op_type_name)
      if attr_def.has_minimum and attr_def.type == "int":
        _SatisfiesIntMinimumConstraint(attr_value.i, attr_def, key,
                                       op_type_name)
      if attr_def.type == "type":
        _SatisfiesTypeConstraint(attr_value.type, attr_def, key)
      if attr_def.type == "list(type)":
        for value in attr_value.list.type:
          _SatisfiesTypeConstraint(value, attr_def, key)

      attr_protos[key] = attr_value
    del attrs  # attrs is no longer authoritative, use attr_protos instead

    # Determine output types (possibly using attrs)
    output_structure = []
    for arg in op_def.output_arg:
      if arg.number_attr:
        n = _AttrValue(attr_protos, arg.number_attr).i
        output_structure.append(n)
      elif arg.type_attr:
        t = _AttrValue(attr_protos, arg.type_attr)
        output_structure.append(None)
      elif arg.type_list_attr:
        t = _AttrValue(attr_protos, arg.type_list_attr)
        output_structure.append(len(t.list.type))
      else:
        output_structure.append(None)

    if keywords:
      raise TypeError("apply_op() got unexpected keyword arguments: " +
                      ", ".join(sorted(keywords.keys())))

    # NOTE(mrry): We add an explicit colocation constraint between
    # the newly created op and any of its reference-typed inputs.
    must_colocate_inputs = [val for arg, val in zip(op_def.input_arg, inputs)
                            if arg.is_ref]
    with _MaybeColocateWith(must_colocate_inputs):
      # Add Op to graph
      # pylint: disable=protected-access
      op = g._create_op_internal(op_type_name, inputs, dtypes=None,
                                 name=scope, input_types=input_types,
                                 attrs=attr_protos, op_def=op_def)

    # `outputs` is returned as a separate return value so that the output
    # tensors can the `op` per se can be decoupled so that the
    # `op_callbacks` can function properly. See framework/op_callbacks.py
    # for more details.
    outputs = op.outputs
    # Conditionally invoke tfdbg v2's op callback(s).
    if op_callbacks.should_invoke_op_callbacks():
      callback_outputs = op_callbacks.invoke_op_callbacks(
          op.node_def.op, tuple(op.inputs), attr_protos, tuple(outputs),
          op_name=op.name, graph=g)
      if callback_outputs is not None:
        outputs = callback_outputs

    return output_structure, op_def.is_stateful, op, outputs


def value_to_attr_value(value, attr_type, arg_name):  # pylint: disable=invalid-name
  """Encodes a Python value as an `AttrValue` proto message.

  Args:
    value: The value to convert.
    attr_type: The value type (string) -- see the AttrValue proto definition for
      valid strings.
    arg_name: Argument name (for error messages).

  Returns:
    An AttrValue proto message that encodes `value`.
  """
  attr_value = attr_value_pb2.AttrValue()

  if attr_type.startswith("list("):
    if not _IsListValue(value):
      raise TypeError("Expected list for attr " + arg_name)

  if attr_type == "string":
    attr_value.s = _MakeStr(value, arg_name)
  elif attr_type == "list(string)":
    attr_value.list.s.extend([_MakeStr(x, arg_name) for x in value])
  elif attr_type == "int":
    attr_value.i = _MakeInt(value, arg_name)
  elif attr_type == "list(int)":
    attr_value.list.i.extend([_MakeInt(x, arg_name) for x in value])
  elif attr_type == "float":
    attr_value.f = _MakeFloat(value, arg_name)
  elif attr_type == "list(float)":
    attr_value.list.f.extend([_MakeFloat(x, arg_name) for x in value])
  elif attr_type == "bool":
    attr_value.b = _MakeBool(value, arg_name)
  elif attr_type == "list(bool)":
    attr_value.list.b.extend([_MakeBool(x, arg_name) for x in value])
  elif attr_type == "type":
    attr_value.type = _MakeType(value, arg_name)
  elif attr_type == "list(type)":
    attr_value.list.type.extend([_MakeType(x, arg_name) for x in value])
  elif attr_type == "shape":
    attr_value.shape.CopyFrom(_MakeShape(value, arg_name))
  elif attr_type == "list(shape)":
    attr_value.list.shape.extend([_MakeShape(x, arg_name) for x in value])
  elif attr_type == "tensor":
    attr_value.tensor.CopyFrom(_MakeTensor(value, arg_name))
  elif attr_type == "list(tensor)":
    attr_value.list.tensor.extend([_MakeTensor(x, arg_name) for x in value])
  elif attr_type == "func":
    attr_value.func.CopyFrom(_MakeFunc(value, arg_name))
  elif attr_type == "list(func)":
    attr_value.list.func.extend([_MakeFunc(x, arg_name) for x in value])
  else:
    raise TypeError("Unrecognized Attr type " + attr_type)
  return attr_value


# The following symbols are used by op_def_util.cc.
_pywrap_utils.RegisterPyObject("tf.dtypes.DType", dtypes.DType)
_pywrap_utils.RegisterPyObject("tf.dtypes.as_dtype", dtypes.as_dtype)
_pywrap_utils.RegisterPyObject("tf.TensorShape", tensor_shape.TensorShape)
_pywrap_utils.RegisterPyObject("tf.as_shape", tensor_shape.as_shape)
_pywrap_utils.RegisterPyObject("tf.TensorProto", tensor_pb2.TensorProto)
_pywrap_utils.RegisterPyObject("text_format.Parse", text_format.Parse)
_pywrap_utils.RegisterPyObject("tf.convert_to_tensor", ops.convert_to_tensor)
